\ 3iMotor.fs - 2020-10-11---Sun---14:51

\ Hardware:
\ 28BYJ-48 5V DC Step Motor, 4 coils, with 3 ways to drive: 
\       wave (A B C D) , full (AB BC CD AD ), or half (A AB B BC C CD D AD)
\       5.625 rotor half-step angle, or 64 steps per revolution
\       The motor has an internal gearbox with 1:63.68395 additional gear reduction => max 4076 microsteps per revolution in half step mode
\           => 0.08833* per micro halfstep  => 12 micro halfsteps per degree (6 steps in wave mode)

\ ULN2003A-PCB motor driver PCB

\ Connections:
\ ports 5 4 3 2 out from Nano to ULN board (2 to IN1)
\ motor to ULN board --- blue to top (aligned with IN1), red at bottom (power)
\ WARNING: power the ULN board separately with 5V from DC volt
\ jumper pin powers the board or cuts power to it -- use a relay to cut power to the motors (they will draw power ambiently)

\ Dec 25, 2018 - 9V to 5V DC converter producing 5V TTL output to 4-button manual step motor driver

: stepper-init  5 4 3 2 port4out ;
: test-port4  16 0 do i port4write 250 ms loop 0 port4write ;

19 constant REG_MOTORCODE

$17 constant CALL_MOTORDEG

4 constant MIN_DELAY  \ fastest is 2 MS, so 1 rev in 4 sec. (i.e. 15 rpm).  4 MS (8 rpm) allows quick changes in rotation direction (with no load),  Top speed is based on gearbox, controller, and driver code.  Faster means less torque.

: deg>> ( delay deg -- )   \ Move motor deg degrees in relative measure, with ms delay between steps (2 is smallest)
    xstore 
    dup 0< if 1 xstore negate else 0 xstore then 
    lohi swap xstore xstore REG_PORTOUT xstore REG_MOTORCODE xstore 
    CALL_MOTORDEG xcall ;
\ ACTION -- can check if motor is done by reading for serial
\ FEATURE -- start by flushing rx queue

: ## MIN_DELAY deg>>  ;

: test-steppers  110 ## -60 ## 40 ## -130 ## 40 ## ;  \ NOTE - these are asynchronous commands, i.e. there is no waiting for rotation to finish    

\ ============== USING 3iMotor primitives deg()


\ ========================   USING 3iForth Primitives only, and building up Motor movements on remote Forth (less efficient) ====================

\ Design A: Use a variable that holds where in the sequence we are, whether stepping forward or backwards.  This means no jitters or loss of position.
\ This is preferred over Design C: c@+, c@-, over Design B: lshift / rshifts (actually should be rotates)

\ Motor drive methods.  
create wave4 1 c, 2 c, 4 c, 8 c,   \ single coil, half power, half torque, 32 steps per rev, 160mA current draw at 5V = 800mW
create full4 3 c, 6 c, 12 c, 9 c,  \ double coil, full power, full torque, 32 steps per rev, 320mA current draw at 5V = 1.6W
create half8 1 c, 3 c, 2 c, 6 c, 4 c, 12 c, 8 c, 9 c,  \ alternating coils, power, and torque, 64 steps per rev 
\ half8 = 4096 steps per revolution (2^12)

\ WAVE MODE, single phase, lowest current, lowest torque
variable mm  \ hold where in the motor code sequence we are, whether stepping forwar or backwards.  Means no jitters or loss of position.
: mm! ( m -- ) mm c! ;  \ store
: mm@ ( -- m ) mm c@ ;  \ fetch
1 mm! \ initialize
: >m  ( idx -- ) wave4 + c@ port4write ;   \ send command to motor
: mm+ ( -- m ) mm@ 1 + 4 mod dup mm! >m ;  \ STEP (rotate) through motor codes, sending instruction to motor
: mm- ( -- m ) mm@ 1 - 4 mod dup mm! >m ;
: spinN ( N -- ) \ 2048 steps per revolution (2^11)  \ spin through steps
    dup 0<> if 
        dup 0> if 0 do mm+ loop else negate 0 do mm- loop then 
    else drop then ;  
28 constant 5degsector  \ 28 steps in 5 deg sector (sectors: 3 (15*), 9 (45*), 18 (90*), 37 (180), 73 (360)
: sectors ( N -- ) 5degsector * spinN ;
: deg ( N -- ) dup 5 * swap 31 * 45 / + spinN ;   \  2048 steps per revolution => 5.6888 steps per degree, or 5 and 2/3 (2040 steps/rev, 0.39% error) or 31/45

: help-motors
cr
cr ." HELP-MOTORS"
cr ." -----------"
cr ."   stepper-init"
cr ."   test-port4  - lights up 4-led output board 0x0 to 0xF"
cr ."   5 sectors   - moves 5x 5-sector degrees using host 3iForth primitives " 
cr ."   2 25 deg>>    25 ##  - moves 25 degrees using remote Motor primitives "
cr ."   test-steppers - operates 5 back and forth moves ASYNCHRONOUSLY (ie without feedback loop/servo)"
cr ;

\ =====================================================
help-motors


\ NEXT STEP:
\ Why does motor get hot?  How to cool it?
\ To go faster, incorporate the spinN code in the AVR
\ Add relay to cut power
\ Measure power consumption
\ Do I really need the ULN PCB, why not just the chip?
\ What about TIP122 transistor, would I need 4 therefore?
\ Why 512 steps in a revolution?  
\ Get a continuous rotate left / rotate right operator...

\ Add HOME and SETPOS functionality
\ Push button and spin fixed amount
\ Press button and keep spinning until release or until second/different press
\ Joystick control
\ Speed control
